/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { CypressConfig } from 'apps/alfa-e2e/cypress.config';
import { readdir, remove } from 'fs-extra';
import { Db, Document, Long, MongoClient, ObjectId, WithId } from 'mongodb';
import { InitOzgCloudFileTaskData, OzgCloudFileE2E } from '../model/binary-file';
import { dropCollections, insert, insertWithDrop } from './mongodb-util';

const fs = require('fs');
const decompress = require('decompress');

const Binary = require('mongodb').Binary;

module.exports = (on: any, config: CypressConfig): CypressConfig => {
  on('task', {
    initCommandData({ collection, data }) {
      console.log('initCommandData');
      insertIntoDatabase(config, collection, parseCommandData(data));
      return 0;
    },
    initGridFsFileData({ collection, data }) {
      console.log('initGridFsFileData');
      insertIntoDatabase(config, collection, parseGridFsFileData(data));
      return 0;
    },
    initGridFsChunkData({ collection, data }) {
      console.log('initGridFsChunkData');
      insertIntoDatabase(config, collection, parseGridFsChunkData(data));
      return 0;
    },
    initOzgCloudFileData(taskData: InitOzgCloudFileTaskData): number {
      console.log('initOzgCloudFileData');
      insertIntoDatabase(config, taskData.collection, parseOzgCloudFileData(taskData.data));
      return 0;
    },
    initVorgangData({ collection, data }) {
      console.log('initVorgangData');
      insertIntoDatabase(config, collection, parseVorgangData(data));
      return 0;
    },
    initVorgangAttachedItemData({ collection, data }) {
      console.log('initVorgangAttachedItemData');
      insertIntoDatabase(config, collection, parseVorgangAttachedItemData(data));
      return 0;
    },
    initUsermanagerData({ collection, data }) {
      console.log('initUsermanagerData');
      insertIntoUserManagerDatabase(config, collection, parseUserData(data));
      return 0;
    },
    dropCollections(collections: string[]): number {
      console.log('dropCollections: ', collections);
      dropCollectionsFromDatabase(config, collections);
      return 0;
    },
    dropUserManagerCollections(collections) {
      console.log('dropUserManagerCollections: ', collections);
      dropUserManagerCollectionsFromDatabase(config, collections);
      return 0;
    },
    countFiles(folderName: string) {
      console.log('counting files in folder %s', folderName);
      return countFiles(folderName);
    },
    deleteFolder(folderName: string) {
      console.log('deleting folder %s', folderName);
      deleteFolder(folderName);
      return 0;
    },
    getDownloadFiles(folderName: string) {
      console.log('get files in folder %s', folderName);
      return getDownloadFiles(folderName);
    },
    unzipDownloadFile({ folderName, fileName }) {
      console.log('unzip file %s in folder %s', fileName, folderName);
      unzipFile(folderName, fileName);
      return 0;
    },
    writeUserIdsToFile({ collection }) {
      writeUserIds(getUserManagerDatabaseUrl(config), getUserManagerDatabase(config), collection).then(
        (userIds: { [username: string]: string }) => {
          fs.writeFileSync('src/fixtures/usermanager/user-ids.json', JSON.stringify(userIds));
        },
      );
      return 0;
    },
  });

  on('after:spec', (spec: Cypress.Spec, results: CypressCommandLine.RunResult) => {
    if (results && results.video && results.stats.failures === 0) {
      console.log(`Delete recorded video because spec passed: ${results.video}`);
      fs.unlinkSync(results.video);
    }
  });

  // Workaround für Angular 13 und Cypress mit Webpack 4,
  // Siehe https://github.com/cypress-io/cypress/issues/19066#issuecomment-1012055705
  // Entfernen, sobald Cypress Webpack 5 nutzt - https://github.com/cypress-io/cypress/issues/19555
  // Ursache: Angular linker needed to link partial-ivy code,
  //   see https://angular.io/guide/creating-libraries#consuming-partial-ivy-code-outside-the-angular-cli
  // Fehlerbild:
  //   - Anwendung läuft im Browser, aber nicht in Cypress.
  //   - Fehlermeldung in Cypress: The injectable 'SystemDateTimeProvider' needs to be compiled using the JIT compiler, but '@angular/compiler' is not available.
  // Lösung:
  //   - NPM-Paket identifizieren, dass "SystemDateTimeProvider" enthält.
  //   - NPM-Paket im "test" Attribut unten hinzufügen.
  const webpackPreprocessor = require('@cypress/webpack-batteries-included-preprocessor');
  const webpackOptions = webpackPreprocessor.defaultOptions.webpackOptions;

  webpackOptions.module.rules.unshift({
    test: /[/\\](@angular|@ngxp|angular-oauth2-oidc)[/\\].+\.m?js$/,
    resolve: {
      fullySpecified: false,
    },
    use: {
      loader: 'babel-loader',
      options: {
        plugins: ['@angular/compiler-cli/linker/babel'],
        compact: false,
        cacheDirectory: true,
      },
    },
  });

  on(
    'file:preprocessor',
    webpackPreprocessor({
      webpackOptions: webpackOptions,
      typescript: require.resolve('typescript'),
    }),
  );

  return config;
  // Ende - Workaround für Angular 13 und Cypress mit Webpack 4
};

function parseCommandData(commands) {
  commands.forEach((command) => parseCommand(command));
  return commands;
}

function parseCommand(command) {
  command.createdAt = createDate(command.createdAt);
  if (command.finishedAt) {
    command.finishedAt = createDate(command.finishedAt);
  }
  command.relationVersion = createNumberLong(command.relationVersion);
}

function parseGridFsFileData(gridFsFiles) {
  gridFsFiles.forEach((gridFsFile) => parseGridFsFile(gridFsFile));
  return gridFsFiles;
}

function parseGridFsFile(gridFsFile) {
  gridFsFile._id = createObjectId(gridFsFile);
  gridFsFile.length = createNumberLong(gridFsFile.length);
  gridFsFile.uploadDate = createDate(gridFsFile.uploadDate);
}

function parseGridFsChunkData(gridFsChunks) {
  gridFsChunks.forEach((gridFsChunk) => parseGridFsChunk(gridFsChunk));
  return gridFsChunks;
}

function parseGridFsChunk(gridFsChunk) {
  gridFsChunk._id = createObjectId(gridFsChunk);
  //TODO createObjectId nutzen, sobald diese umgestellt ist
  gridFsChunk.files_id = new ObjectId(gridFsChunk.files_id.$oid);
  gridFsChunk.data = createBinData(gridFsChunk.data);
}

function parseOzgCloudFileData(ozgCloudFiles: OzgCloudFileE2E[]) {
  ozgCloudFiles.forEach((ozgCloudFile: OzgCloudFileE2E) => parseOzgCloudFile(ozgCloudFile));
  return ozgCloudFiles;
}

function parseOzgCloudFile(ozgCloudFile) {
  ozgCloudFile._id = createObjectId(ozgCloudFile);
  ozgCloudFile.size = createNumberLong(ozgCloudFile.size);
  ozgCloudFile.version = createNumberLong(ozgCloudFile.version);
}

function parseVorgangData(data) {
  data.forEach((vorgang) => parseVorgang(vorgang));
  return data;
}

function parseVorgang(vorgang) {
  vorgang._id = createObjectId(vorgang);
  vorgang.createdAt = createDate(vorgang.createdAt);

  vorgang.eingangs.forEach((eingang) => parseEingang(eingang));

  if (vorgang.wiedervorlages) {
    vorgang.wiedervorlages.forEach((wiedervorlage) => parseWiedervorlage(wiedervorlage));
  }
  if (vorgang.kommentars) {
    vorgang.kommentars.forEach((kommentar) => parseKommentar(kommentar));
  }
}

function parseEingang(eingang) {
  eingang.header.createdAt = createDate(eingang.header.createdAt);
}

function createBinData(encoded64Value) {
  return Binary(Buffer.from(encoded64Value, 'base64'), 0);
}

function parseWiedervorlage(wiedervorlage) {
  wiedervorlage.frist = createDate(wiedervorlage.frist);
  wiedervorlage.createdAt = createDate(wiedervorlage.createdAt);
}

function parseKommentar(kommentar) {
  kommentar.createdAt = createDate(kommentar.createdAt);
}

function createDate(field) {
  return new Date(field.$date);
}

function parseVorgangAttachedItemData(vorgangAttachedItems) {
  vorgangAttachedItems.forEach((vorgangAttachedItem) => parseVorgangAttachedItem(vorgangAttachedItem));
  return vorgangAttachedItems;
}

function parseVorgangAttachedItem(parseVorgangAttachedItem) {
  parseVorgangAttachedItem._id = createObjectId(parseVorgangAttachedItem);
  parseVorgangAttachedItem.version = createNumberLong(parseVorgangAttachedItem.version);
}

function parseUserData(data) {
  data.forEach((user) => parseUser(user));
  return data;
}

function parseUser(user) {
  user._id = createObjectId(user);
  user.createdAt = createDate(user.createdAt);
}

//TOOD Beschraenkung auf _id aufheben
function createObjectId(field) {
  return new ObjectId(field._id.$oid);
}

function createNumberLong(numberValue: number): Long {
  return Long.fromNumber(numberValue);
}

function insertIntoDatabase(config: CypressConfig, collection: string, data): void {
  insertWithDrop(getDatabaseUrl(config), getDatabase(config), collection, data);
}

function insertIntoUserManagerDatabase(config: CypressConfig, collection: string, data): void {
  insert(getUserManagerDatabaseUrl(config), getUserManagerDatabase(config), collection, data);
}

function dropCollectionsFromDatabase(config: CypressConfig, collections: string[]): void {
  dropCollections(getDatabaseUrl(config), getDatabase(config), collections);
}

function getDatabaseUrl(config: CypressConfig): string {
  return config.env.dbUrl;
}

function getDatabase(config: CypressConfig): string {
  return config.env.database;
}

function dropUserManagerCollectionsFromDatabase(config: CypressConfig, collections) {
  dropCollections(getUserManagerDatabaseUrl(config), getUserManagerDatabase(config), collections);
}

function getUserManagerDatabaseUrl(config: CypressConfig): string {
  return config.env.userManager.dbUrl;
}

function getUserManagerDatabase(config: CypressConfig): string {
  return config.env.userManager.database;
}

function countFiles(folderName) {
  return readdir(folderName)
    .then((files) => files.length)
    .catch((err) => {
      throw new Error(`Counting files failed: ${err.message}`);
    });
}

function deleteFolder(folderName: string): void {
  return remove(folderName)
    .then(() => {
      return null;
    })
    .catch((err) => {
      throw new Error(`Failed to delete folder: ${err.message}`);
    });
}

function getDownloadFiles(folderName: string): Promise<Array<string>> {
  return fs.readdirSync(folderName);
}

function unzipFile(folderName: string, fileName: string): void {
  console.log('File: ' + folderName + fileName);

  decompress(folderName + '/' + fileName, folderName);
  return null;
}

async function writeUserIds(databaseUrl, databaseName, collection) {
  return MongoClient.connect(databaseUrl).then((client: MongoClient) => {
    const db: Db = client.db(databaseName);
    const userIds: { [username: string]: string } = {};
    return db
      .collection(collection)
      .find()
      .map((doc: WithId<Document>) => [doc['username'], doc._id.toString()])
      .toArray()
      .then((docData: [string, string][]) => {
        docData.forEach((data) => Object.assign(userIds, { [data[0]]: data[1] }));
        return userIds;
      });
  });
}
