import { mockWindowError } from '@alfa-client/test-utils';
import { FormControl, FormGroup } from '@angular/forms';
import { patchForm } from './form.util';

describe('FormUtil', () => {
  describe('patch form', () => {
    it('should not throw any errors', () => {
      const errorSpy: jest.SpyInstance = mockWindowError();
      const formGroup: FormGroup = new FormGroup({
        existingKey: new FormControl(null),
      });

      patchForm({ missingKey: 'dummyValue' }, formGroup);

      expect(errorSpy).not.toHaveBeenCalled();
    });

    describe('on strings', () => {
      it('should patch values', () => {
        const formGroup: FormGroup = new FormGroup({
          stringValue: new FormControl(null),
        });

        patchForm({ stringValue: 'test' }, formGroup);

        expect(formGroup.controls['stringValue'].value).toBe('test');
      });
    });

    describe('on array string value', () => {
      it('should patch all', () => {
        const formGroup: FormGroup = new FormGroup({
          array: new FormGroup({
            arrayValue1: new FormControl(false),
            arrayValue2: new FormControl(false),
          }),
        });

        patchForm({ array: ['arrayValue1', 'arrayValue2'] }, formGroup);

        const arrayFormGroup: FormGroup = getControlAsFormGroup('array', formGroup);
        expect(arrayFormGroup.controls['arrayValue1'].value).toBeTruthy();
        expect(arrayFormGroup.controls['arrayValue2'].value).toBeTruthy();
      });

      it('should patch matching values only', () => {
        const formGroup: FormGroup = new FormGroup({
          array: new FormGroup({
            arrayValue1: new FormControl(false),
            arrayValue2: new FormControl(false),
          }),
        });

        patchForm({ array: ['arrayValue1'] }, formGroup);

        const arrayFormGroup: FormGroup = getControlAsFormGroup('array', formGroup);
        expect(arrayFormGroup.controls['arrayValue1'].value).toBeTruthy();
        expect(arrayFormGroup.controls['arrayValue2'].value).toBeFalsy();
      });

      it('should not throw any error if control is missing', () => {
        const errorSpy: jest.SpyInstance = mockWindowError();
        const formGroup: FormGroup = new FormGroup({ array: new FormGroup({}) });

        patchForm({ array: ['arrayValue1'] }, formGroup);

        expect(errorSpy).not.toHaveBeenCalled();
      });
    });

    describe('on object value', () => {
      describe('contains strings', () => {
        it('should patch values', () => {
          const formGroup: FormGroup = new FormGroup({
            object: new FormGroup({
              stringValue: new FormControl(null),
            }),
          });

          patchForm({ object: { stringValue: 'test' } }, formGroup);

          const objectFormGroup: FormGroup = getControlAsFormGroup('object', formGroup);
          expect(objectFormGroup.controls['stringValue'].value).toEqual('test');
        });
      });

      describe('contains array string value', () => {
        it('should patch all', () => {
          const formGroup: FormGroup = new FormGroup({
            object: new FormGroup({
              array: new FormGroup({
                arrayValue1: new FormControl(false),
                arrayValue2: new FormControl(false),
              }),
            }),
          });

          patchForm({ object: { array: ['arrayValue1', 'arrayValue2'] } }, formGroup);

          const objectFormGroup: FormGroup = getControlAsFormGroup('object', formGroup);
          const arrayFormGroup: FormGroup = getControlAsFormGroup('array', objectFormGroup);
          expect(arrayFormGroup.controls['arrayValue1'].value).toBeTruthy();
          expect(arrayFormGroup.controls['arrayValue2'].value).toBeTruthy();
        });

        it('should patch matching values only', () => {
          const formGroup: FormGroup = new FormGroup({
            object: new FormGroup({
              array: new FormGroup({
                arrayValue1: new FormControl(false),
                arrayValue2: new FormControl(false),
              }),
            }),
          });

          patchForm({ object: { array: ['arrayValue1'] } }, formGroup);

          const objectFormGroup: FormGroup = getControlAsFormGroup('object', formGroup);
          const arrayFormGroup: FormGroup = getControlAsFormGroup('array', objectFormGroup);
          expect(arrayFormGroup.controls['arrayValue1'].value).toBeTruthy();
          expect(arrayFormGroup.controls['arrayValue2'].value).toBeFalsy();
        });
      });
    });
  });

  function getControlAsFormGroup(key: string, formGroup: FormGroup): FormGroup {
    return <FormGroup>formGroup.controls[key];
  }
});
