import { ROUTES } from '@admin-client/shared';
import { UserService } from '@admin-client/user-shared';
import { NavigationService } from '@alfa-client/navigation-shared';
import { createEmptyStateResource, StateResource } from '@alfa-client/tech-shared';
import { dispatchEventFromFixture, getMockComponent, Mock, mock } from '@alfa-client/test-utils';
import { ComponentFixture, TestBed } from '@angular/core/testing';
import { faker } from '@faker-js/faker/locale/de';
import { Observable, of } from 'rxjs';
import { getDataTestIdOf } from '../../../../../../tech-shared/test/data-test';
import { singleColdCompleted } from '../../../../../../tech-shared/test/marbles';
import { createUserFormGroup } from '../../../../test/form';
import { UserFormService } from '../user.formservice';
import { UserDeleteDialogContainerComponent } from './user-delete-dialog-container.component';
import { UserDeleteDialogComponent } from './user-delete-dialog/user-delete-dialog.component';

describe('UserDeleteDialogContainerComponent', () => {
  let component: UserDeleteDialogContainerComponent;
  let fixture: ComponentFixture<UserDeleteDialogContainerComponent>;

  const deletDialogLocator: string = getDataTestIdOf('delete-dialog');

  let formService: Mock<UserFormService>;
  let userService: Mock<UserService>;
  let navigationService: Mock<NavigationService>;

  beforeEach(async () => {
    formService = mock(UserFormService);
    userService = mock(UserService);
    navigationService = mock(NavigationService);
    formService = {
      ...mock(UserFormService),
      form: createUserFormGroup(),
      delete: jest.fn(),
    } as any;

    await TestBed.configureTestingModule({
      imports: [UserDeleteDialogContainerComponent, UserDeleteDialogComponent],
      providers: [
        { provide: UserFormService, useValue: formService },
        { provide: UserService, useValue: userService },
        { provide: NavigationService, useValue: navigationService },
      ],
    }).compileComponents();

    fixture = TestBed.createComponent(UserDeleteDialogContainerComponent);
    component = fixture.componentInstance;
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  describe('user form delete dialog', () => {
    it('should be called', () => {
      const username: string = faker.word.sample();
      const stateResource: StateResource<unknown> = createEmptyStateResource();
      formService.getUserName = jest.fn().mockReturnValue(username);
      component.deleteUserStateResource$ = of(stateResource);

      fixture.detectChanges();

      const deleteDialog: UserDeleteDialogComponent = getMockComponent(fixture, UserDeleteDialogComponent);
      expect(deleteDialog.username).toBe(username);
      expect(deleteDialog.deleteUserStateResource).toBe(stateResource);
    });

    it('should call deleteUser on emit delete', () => {
      component.deleteUser = jest.fn();

      dispatchEventFromFixture(fixture, deletDialogLocator, 'delete');

      expect(component.deleteUser).toHaveBeenCalled();
    });
  });

  describe('component', () => {
    describe('deleteUser', () => {
      const userId: string = faker.string.uuid();
      const loadingStateResource: StateResource<unknown> = createEmptyStateResource(true);
      const loadingStateResource$: Observable<StateResource<unknown>> = of(loadingStateResource);
      const loadingDoneStateResource$: Observable<StateResource<unknown>> = of(createEmptyStateResource());

      beforeEach(() => {
        userService.delete = jest.fn().mockReturnValue(loadingStateResource$);
        formService.getId = jest.fn().mockReturnValue(userId);
      });

      it('should call formService getId', () => {
        component.deleteUser();

        expect(formService.getId).toHaveBeenCalled();
      });

      it('should call userService delete', () => {
        component.deleteUser();

        expect(userService.delete).toHaveBeenCalledWith(userId);
      });

      it('should set deleteUserStateResource$', () => {
        component.deleteUser();

        expect(component.deleteUserStateResource$).toBeObservable(singleColdCompleted(loadingStateResource));
      });

      it('should not navigate on loading', () => {
        component.deleteUser();
        component.deleteUserStateResource$.subscribe();

        expect(navigationService.navigate).not.toHaveBeenCalled();
      });

      it('should navigate on delete success', () => {
        userService.delete.mockReturnValue(loadingDoneStateResource$);

        component.deleteUser();
        component.deleteUserStateResource$.subscribe();

        expect(navigationService.navigate).toHaveBeenCalledWith(ROUTES.BENUTZER);
      });
    });
  });
});
