/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { AdminOrganisationsEinheit, AdminOrganisationsEinheitService } from '@admin-client/organisations-einheit-shared';
import { ROUTES } from '@admin-client/shared';
import { User, UserService } from '@admin-client/user-shared';
import { PatchConfig } from '@admin/keycloak-shared';
import { NavigationService } from '@alfa-client/navigation-shared';
import { createEmptyStateResource, createLoadingStateResource, createStateResource, StateResource, } from '@alfa-client/tech-shared';
import { Mock, mock, mockWindowError } from '@alfa-client/test-utils';
import { SnackBarService } from '@alfa-client/ui';
import { fakeAsync, TestBed, tick } from '@angular/core/testing';
import { AbstractControl, FormControl, FormGroup, UntypedFormBuilder, UntypedFormGroup } from '@angular/forms';
import { ActivatedRoute, UrlSegment } from '@angular/router';
import { faker } from '@faker-js/faker/locale/de';
import { expect } from '@jest/globals';
import { cold } from 'jest-marbles';
import { createUser } from 'libs/admin/user-shared/test/user';
import { Observable, of, throwError } from 'rxjs';
import { createUrlSegment } from '../../../../../navigation-shared/test/navigation-test-factory';
import { singleCold, singleColdCompleted, singleHot } from '../../../../../tech-shared/test/marbles';
import { createAdminOrganisationsEinheit } from '../../../../organisations-einheit-shared/src/test/organisations-einheit';
import { UserFormService } from './user.formservice';
import SpyInstance = jest.SpyInstance;

describe('UserFormService', () => {
  let formService: UserFormService;

  let roleGroup: UntypedFormGroup;
  let alfaGroup: UntypedFormGroup;
  let organisationsEinheitenGroup: UntypedFormGroup;
  let administrationGroup: UntypedFormGroup;

  let service: Mock<UserService>;
  let adminOrganisationsEinheitService: Mock<AdminOrganisationsEinheitService>;
  let navigationService: Mock<NavigationService>;
  let snackBarService: Mock<SnackBarService>;
  let activatedRoute: Mock<ActivatedRoute>;

  const adminOrganisationsEinheit: AdminOrganisationsEinheit = createAdminOrganisationsEinheit();
  const adminOrganisationsEinheitList: StateResource<AdminOrganisationsEinheit[]> = createStateResource([
    adminOrganisationsEinheit,
  ]);

  beforeEach(() => {
    service = {
      ...mock(UserService),
      refresh: jest.fn(),
      create: jest.fn(),
      save: jest.fn(),
      delete: jest.fn(),
      getUserById: jest.fn(),
    };
    adminOrganisationsEinheitService = {
      ...mock(AdminOrganisationsEinheitService),
      getAll: jest.fn().mockReturnValue(of(adminOrganisationsEinheitList)),
    };
    navigationService = mock(NavigationService);
    snackBarService = mock(SnackBarService);
    activatedRoute = <any>{ ...mock(ActivatedRoute), url: of<UrlSegment[]>([]) };

    TestBed.configureTestingModule({
      providers: [
        UserFormService,
        UntypedFormBuilder,
        { provide: UserService, useValue: service },
        { provide: AdminOrganisationsEinheitService, useValue: adminOrganisationsEinheitService },
        { provide: NavigationService, useValue: navigationService },
        { provide: SnackBarService, useValue: snackBarService },
        { provide: ActivatedRoute, useValue: activatedRoute },
      ],
    });

    formService = TestBed.inject(UserFormService);
    organisationsEinheitenGroup = <UntypedFormGroup>formService.form.get(UserFormService.GROUPS);
    roleGroup = <UntypedFormGroup>formService.form.get(UserFormService.CLIENT_ROLES);
    alfaGroup = <UntypedFormGroup>roleGroup.get(UserFormService.ALFA_GROUP);
    administrationGroup = <UntypedFormGroup>roleGroup.get(UserFormService.ADMINISTRATION_GROUP);
  });

  it('should create', () => {
    expect(formService).toBeTruthy();
  });

  describe('build patch config', () => {
    describe('on matching route', () => {
      it('should contains id and value for patch indication', () => {
        const benutzerUrlSegment: UrlSegment = <UrlSegment>{ ...createUrlSegment(), path: 'benutzer' };
        const idUrlSegment: UrlSegment = <UrlSegment>{ ...createUrlSegment(), path: 'dummyId' };

        const patchConfig: PatchConfig = formService._buildPatchConfig([benutzerUrlSegment, idUrlSegment]);

        expect(patchConfig).toEqual({ id: 'dummyId', doPatch: true });
      });
    });

    it('should contains value for patch indication only on non matching route', () => {
      const benutzerUrlSegment: UrlSegment = <UrlSegment>{ ...createUrlSegment(), path: 'benutzer' };
      const idUrlSegment: UrlSegment = <UrlSegment>{ ...createUrlSegment(), path: 'neu' };

      const patchConfig: PatchConfig = formService._buildPatchConfig([benutzerUrlSegment, idUrlSegment]);

      expect(patchConfig).toEqual({ doPatch: false });
    });
  });

  describe('load', () => {
    const id: string = 'dummyId';

    const loadedUser: StateResource<User> = createStateResource(createUser());

    beforeEach(() => {
      service.getUserById.mockReturnValue(singleHot(loadedUser));
    });

    it('should call service to get user by id', () => {
      formService._load(id);

      expect(service.getUserById).toHaveBeenCalledWith(id);
    });

    it('should return loaded user', () => {
      const response: Observable<StateResource<User>> = formService._load(id);

      expect(response).toBeObservable(singleCold(loadedUser));
    });
  });

  describe('listenToAlfaGroupChanges', () => {
    it('should call handleAlfaGroupChange on initial change', () => {
      formService._handleAlfaGroupChange = jest.fn();

      formService.listenToAlfaGroupChanges();

      expect(formService._handleAlfaGroupChange).toHaveBeenCalled();
    });

    it('should call handleAlfaGroupChange when value of form element changes', fakeAsync(() => {
      formService._handleAlfaGroupChange = jest.fn();

      alfaGroup.get(UserFormService.LOESCHEN).setValue(true);

      tick();

      expect(formService._handleAlfaGroupChange).toHaveBeenCalled();
    }));
  });

  describe('initOrganisationsEinheiten', () => {
    beforeEach(() => {
      formService._addOrganisationsEinheitenToForm = jest.fn();
    });

    it('should call adminOrganisationsEinheitService getAll', () => {
      formService._initOrganisationsEinheiten();

      expect(adminOrganisationsEinheitService.getAll).toHaveBeenCalled();
    });

    it('should return resource', () => {
      const result: Observable<AdminOrganisationsEinheit[]> = formService._initOrganisationsEinheiten();

      expect(result).toBeObservable(singleColdCompleted(adminOrganisationsEinheitList.resource));
    });

    it('should call addOrganisationsEinheitenToForm ', fakeAsync(() => {
      formService._initOrganisationsEinheiten().subscribe();
      tick();

      expect(formService._addOrganisationsEinheitenToForm).toHaveBeenCalled();
    }));

    it('should save ids in map', () => {
      formService._initOrganisationsEinheiten().subscribe();

      expect(formService._organisationsEinheitToGroupIdMap.get(adminOrganisationsEinheit.name)).toEqual(
        adminOrganisationsEinheit.id,
      );
    });

    it('should set initOrganisationsEinheiten$', () => {
      expect(formService['_initOrganisationsEinheiten$']).toBeDefined();
    });

    it('should not throw any exception on loading state resource', () => {
      adminOrganisationsEinheitService.getAll.mockReturnValue(of(createLoadingStateResource()));
      const errorMock: any = mockWindowError();

      formService._initOrganisationsEinheiten();

      expect(errorMock).not.toHaveBeenCalled();
    });
  });

  describe('addOrganisationsEinheitenToForm', () => {
    it('should add organisationsEinheiten to form', () => {
      const organisationsEinheitenGroup: UntypedFormGroup = <UntypedFormGroup>formService.form.get(UserFormService.GROUPS);

      formService._addOrganisationsEinheitenToForm(adminOrganisationsEinheitList.resource, organisationsEinheitenGroup);

      expect(organisationsEinheitenGroup.value).toEqual({ [adminOrganisationsEinheit.name]: false });
    });
  });

  describe('roleValidator', () => {
    it('should return error if no role is selected', () => {
      const result = formService.roleValidator()(roleGroup);

      expect(result).toEqual({ atLeastOneRoleSelected: true });
    });

    it('should return null if at least one role is selected', () => {
      alfaGroup.get(UserFormService.LOESCHEN).setValue(true);

      const result = formService.roleValidator()(roleGroup);

      expect(result).toBeNull();
    });
  });

  describe('handleAlfaGroupChange', () => {
    it('should call disableUncheckedCheckboxes if any checkbox is checked', () => {
      formService._isAnyChecked = jest.fn().mockReturnValue(true);
      formService._disableUncheckedCheckboxes = jest.fn();

      formService._handleAlfaGroupChange(alfaGroup);

      expect(formService._disableUncheckedCheckboxes).toHaveBeenCalled();
    });

    it('should call enableAllCheckboxes if not any checkbox is checked', () => {
      formService._isAnyChecked = jest.fn().mockReturnValue(false);
      formService._enableAllCheckboxes = jest.fn();

      formService._handleAlfaGroupChange(alfaGroup);

      expect(formService._enableAllCheckboxes).toHaveBeenCalled();
    });
  });

  describe('isAnyChecked', () => {
    it('should return false if no checkbox is checked', () => {
      const result = formService._isAnyChecked(alfaGroup);

      expect(result).toBe(false);
    });

    it('should return true if any checkbox is checked', () => {
      alfaGroup.get(UserFormService.LOESCHEN).setValue(true);

      const result = formService._isAnyChecked(alfaGroup);

      expect(result).toBe(true);
    });
  });

  describe('disableUncheckedCheckboxes', () => {
    it('if control value is false then control should be disabled', () => {
      const control: AbstractControl = alfaGroup.get(UserFormService.LOESCHEN);
      control.setValue(false);

      formService._disableUncheckedCheckboxes(alfaGroup);

      expect(control.disabled).toBe(true);
    });

    it('if control value is true then control should NOT be disabled', () => {
      const control: AbstractControl = alfaGroup.get(UserFormService.LOESCHEN);
      control.setValue(true);

      formService._disableUncheckedCheckboxes(alfaGroup);

      expect(control.disabled).toBe(false);
    });
  });

  describe('updateCheckboxStates', () => {
    it('if control value is false then control should be disabled', () => {
      const control: AbstractControl = alfaGroup.get(UserFormService.LOESCHEN);
      control.setValue(false);

      formService._disableUncheckedCheckboxes(alfaGroup);

      expect(control.disabled).toBe(true);
    });
  });

  describe('enableAllCheckboxes', () => {
    it('if control value is true then control should be enabled', () => {
      const control: AbstractControl = alfaGroup.get(UserFormService.LOESCHEN);
      const enableSpy = jest.spyOn(control, 'enable');

      formService._enableAllCheckboxes(alfaGroup);

      expect(enableSpy).toHaveBeenCalled();
    });
  });

  describe('doSubmit', () => {
    const user: User = createUser();

    beforeEach(() => {
      formService.isInvalid = jest.fn().mockReturnValue(false);
      service.create.mockReturnValue(of(createStateResource(user)));
    });

    it('should return empty stateResource if form is invalid', () => {
      formService.isInvalid = jest.fn().mockReturnValue(true);

      const result: Observable<StateResource<User>> = formService._doSubmit();

      expect(result).toBeObservable(singleColdCompleted(createEmptyStateResource<User>()));
    });

    it('should call createUser', () => {
      formService._createUser = jest.fn();

      formService.submit();

      expect(formService._createUser).toHaveBeenCalled();
    });

    it('should call _createOrSave', () => {
      formService._createUser = jest.fn().mockReturnValue(user);
      formService._createOrSave = jest.fn().mockReturnValue(of(user));

      formService.submit();

      expect(formService._createOrSave).toHaveBeenCalledWith(user);
    });

    it('should call handleOnCreateUserSuccess if not loading', fakeAsync(() => {
      const handleOnCreateUserSuccessSpy: SpyInstance = jest.spyOn(formService, 'handleOnCreateUserSuccess');

      formService.submit().subscribe();
      tick();

      expect(handleOnCreateUserSuccessSpy).toHaveBeenCalled();
    }));

    it('should not call handleOnCreateUserSuccess if loading', fakeAsync(() => {
      service.create.mockReturnValue(of(createEmptyStateResource(true)));
      const handleOnCreateUserSuccessSpy: SpyInstance = jest.spyOn(formService, 'handleOnCreateUserSuccess');

      formService.submit().subscribe();
      tick();

      expect(handleOnCreateUserSuccessSpy).not.toHaveBeenCalled();
    }));

    it('should call handleSubmitError on error', fakeAsync(() => {
      service.create.mockReturnValue(throwError(() => new Error()));
      const handleSubmitErrorSpy: SpyInstance = jest.spyOn(formService, 'handleSubmitError');

      formService.submit().subscribe();
      tick();

      expect(handleSubmitErrorSpy).toHaveBeenCalled();
    }));
  });

  describe('createOrSave', () => {
    const user: User = createUser();

    const id: string = faker.string.uuid();

    const patchConfig: PatchConfig = { id, doPatch: true };

    it('should call save if patched', () => {
      formService.isPatch = jest.fn().mockReturnValue(true);
      formService._patchConfig = patchConfig;
      user.groups = [adminOrganisationsEinheit.name];
      formService._organisationsEinheitToGroupIdMap.set(adminOrganisationsEinheit.name, adminOrganisationsEinheit.id);

      formService._createOrSave(user);

      expect(service.save).toHaveBeenCalledWith({ ...user, id, groupIds: [adminOrganisationsEinheit.id] });
    });

    it('should call create if not patched', () => {
      formService.isPatch = jest.fn().mockReturnValue(false);

      formService._createOrSave(user);

      expect(service.create).toHaveBeenCalledWith(user);
    });
  });

  describe('handleOnCreateUserSuccess', () => {
    it('should show success message for create', () => {
      formService.handleOnCreateUserSuccess();

      expect(snackBarService.showInfo).toHaveBeenCalledWith('Der Benutzer wurde angelegt.');
    });

    it('should show success message for patch', () => {
      formService.isPatch = jest.fn().mockReturnValue(true);

      formService.handleOnCreateUserSuccess();

      expect(snackBarService.showInfo).toHaveBeenCalledWith('Der Benutzer wurde gespeichert.');
    });

    it('should navigate back to user list', () => {
      formService.handleOnCreateUserSuccess();

      expect(navigationService.navigate).toHaveBeenCalledWith(ROUTES.BENUTZER);
    });
  });

  describe('handleSubmitError', () => {
    it('should return empty stateResource', () => {
      const result = formService.handleSubmitError();

      expect(result).toBeObservable(cold('(a|)', { a: createEmptyStateResource<User>() }));
    });

    it('should show error message', fakeAsync(() => {
      formService.handleSubmitError().subscribe();
      tick();

      expect(snackBarService.showError).toHaveBeenCalledWith('Der Benutzer konnte nicht angelegt werden.');
    }));
  });

  describe('getRoles', () => {
    it('should return no roles when none are active', () => {
      const result: string[] = formService._getRoles(UserFormService.ALFA_GROUP);

      expect(result).toEqual([]);
    });

    it('should return poststelle role when active', () => {
      alfaGroup.get(UserFormService.POSTSTELLE).setValue(true);

      const result: string[] = formService._getRoles(UserFormService.ALFA_GROUP);

      expect(result).toEqual([UserFormService.POSTSTELLE]);
    });
  });

  describe('getActiveOrganisationsEinheiten', () => {
    it('should return no groups when none are active', () => {
      const result: string[] = formService._getActiveOrganisationsEinheiten();

      expect(result).toEqual([]);
    });

    it('should return bauamt group when active', () => {
      const organisationsEinheit: AdminOrganisationsEinheit = createAdminOrganisationsEinheit();
      organisationsEinheitenGroup.addControl(organisationsEinheit.name, new FormControl(true));

      const result: string[] = formService._getActiveOrganisationsEinheiten();

      expect(result).toEqual([organisationsEinheit.name]);
    });
  });

  describe('ngOnDestroy', () => {
    it('should unsubscribe from initOrganisationsEinheiten$', () => {
      formService._initOrganisationsEinheiten$.unsubscribe = jest.fn();

      formService.ngOnDestroy();

      expect(formService._initOrganisationsEinheiten$.unsubscribe).toHaveBeenCalled();
    });

    it('should unsubscribe from initOrganisationsEinheiten$', () => {
      formService._alfaGroupChanges.unsubscribe = jest.fn();

      formService.ngOnDestroy();

      expect(formService._alfaGroupChanges.unsubscribe).toHaveBeenCalled();
    });
  });

  describe('get userName', () => {
    it('should return form control value of userName', () => {
      formService.form = new FormGroup({ [UserFormService.USERNAME]: new FormControl('userNameDummy') });

      const userName: string = formService.getUserName();

      expect(userName).toBe('userNameDummy');
    });
  });
});
