/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { CollaborationListResource } from '@alfa-client/collaboration-shared';
import { CommandResource, hasCommandError, tapOnCommandSuccessfullyDone } from '@alfa-client/command-shared';
import { AbstractFormService, StateResource } from '@alfa-client/tech-shared';
import { SnackBarService } from '@alfa-client/ui';
import { Injectable } from '@angular/core';
import { FormBuilder, FormControl, FormGroup } from '@angular/forms';
import { ResourceUri } from '@ngxp/rest';
import { CollaborationService } from 'libs/collaboration-shared/src/lib/collaboration.service';
import { isNil } from 'lodash-es';
import { Observable, tap } from 'rxjs';

@Injectable()
export class CollaborationRequestFormService extends AbstractFormService<CommandResource> {
  public static readonly FIELD_ZUSTAENDIGE_STELLE: string = 'zustaendigeStelle';
  public static readonly FIELD_TITEL: string = 'titel';
  public static readonly FIELD_BESCHREIBUNG: string = 'beschreibung';

  public static readonly GENERIC_ERROR_MESSAGE: string = 'Zuarbeit konnte nicht angefragt werden';

  private static readonly PATH_PREFIX: string = 'command.body';

  listResource: CollaborationListResource;

  constructor(
    formBuilder: FormBuilder,
    private service: CollaborationService,
    private snackBarService: SnackBarService,
  ) {
    super(formBuilder);
  }

  protected initForm(): FormGroup {
    return this.formBuilder.group({
      [CollaborationRequestFormService.FIELD_ZUSTAENDIGE_STELLE]: new FormControl<ResourceUri>(null),
      [CollaborationRequestFormService.FIELD_TITEL]: new FormControl<string>(null),
      [CollaborationRequestFormService.FIELD_BESCHREIBUNG]: new FormControl<string>(null),
    });
  }

  protected doSubmit(): Observable<StateResource<CommandResource>> {
    return this.service.create(this.listResource, this.getFormValue()).pipe(
      tapOnCommandSuccessfullyDone(() => this.service.refreshList()),
      tap((commandStateResource: StateResource<CommandResource>) => {
        this.handleCommandError(commandStateResource.resource);
      }),
    );
  }

  protected handleCommandError(commandResource: CommandResource): void {
    if (isNil(commandResource)) return;

    if (hasCommandError(commandResource)) {
      this.showError();
    }
  }

  protected showError(): void {
    this.snackBarService.showError(CollaborationRequestFormService.GENERIC_ERROR_MESSAGE);
  }

  protected getPathPrefix(): string {
    return CollaborationRequestFormService.PATH_PREFIX;
  }

  public setListResource(listResource: CollaborationListResource): void {
    this.listResource = listResource;
  }
}
