/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { BinaryFileAttachmentContainerComponent } from '@alfa-client/binary-file';
import { CommandResource } from '@alfa-client/command-shared';
import { KommentarLinkRel, KommentarListResource, KommentarService } from '@alfa-client/kommentar-shared';
import { createEmptyStateResource, createErrorStateResource, createStateResource, StateResource } from '@alfa-client/tech-shared';
import { existsAsHtmlElement, Mock, mock, triggerEvent, useFromMock } from '@alfa-client/test-utils';
import { OzgcloudStrokedButtonWithSpinnerComponent, TextAreaEditorComponent } from '@alfa-client/ui';
import { ComponentFixture, TestBed } from '@angular/core/testing';
import { FormBuilder, ReactiveFormsModule } from '@angular/forms';
import { MatFormFieldModule } from '@angular/material/form-field';
import { expect } from '@jest/globals';
import { MockComponent } from 'ng-mocks';
import { EMPTY, of } from 'rxjs';
import { createBinaryFileListResource } from '../../../../../binary-file-shared/test/binary-file';
import { MultiFileUploadComponent } from '../../../../../binary-file/src/lib/multi-file-upload/multi-file-upload.component';
import { createSuccessfullyDoneCommandStateResource } from '../../../../../command-shared/test/command';
import { createKommentarListResource, createKommentarResource } from '../../../../../kommentar-shared/test/kommentar';
import { getDataTestIdOf } from '../../../../../tech-shared/test/data-test';
import { createProblemDetail } from '../../../../../tech-shared/test/error';
import { singleColdCompleted } from '../../../../../tech-shared/test/marbles';
import { KommentarFormComponent } from './kommentar-form.component';
import { KommentarFormService } from './kommentar.formservice';

describe('KommentarFormComponent', () => {
  let component: KommentarFormComponent;
  let fixture: ComponentFixture<KommentarFormComponent>;

  const cancelButtonTestId: string = getDataTestIdOf('cancel-button');

  const kommentarListStateResource: StateResource<KommentarListResource> = createStateResource(createKommentarListResource());

  let kommentarService: Mock<KommentarService>;
  let formService: KommentarFormService;

  beforeEach(() => {
    formService = new KommentarFormService(new FormBuilder(), useFromMock(kommentarService));
    kommentarService = mock(KommentarService);
  });

  beforeEach(async () => {
    await TestBed.configureTestingModule({
      declarations: [
        KommentarFormComponent,
        MockComponent(TextAreaEditorComponent),
        MockComponent(OzgcloudStrokedButtonWithSpinnerComponent),
        MockComponent(BinaryFileAttachmentContainerComponent),
        MockComponent(MultiFileUploadComponent),
      ],
      imports: [MatFormFieldModule, ReactiveFormsModule],
      providers: [
        {
          provide: KommentarService,
          useValue: kommentarService,
        },
      ],
    })
      .overrideComponent(KommentarFormComponent, {
        set: {
          providers: [
            {
              provide: KommentarFormService,
              useValue: formService,
            },
          ],
        },
      })
      .compileComponents();

    fixture = TestBed.createComponent(KommentarFormComponent);
    component = fixture.componentInstance;
    component.kommentarListStateResource = kommentarListStateResource;
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  it('should call patch', () => {
    const patchSpy = jest.spyOn(KommentarFormService.prototype, 'patch').mockImplementation();
    const kommentarResource = createKommentarResource([KommentarLinkRel.ATTACHMENTS]);
    component.kommentar = kommentarResource;
    kommentarService.getAttachments.mockReturnValue(of(createStateResource(createBinaryFileListResource())));

    component.ngOnChanges();

    expect(patchSpy).toHaveBeenCalledWith(kommentarResource);
  });

  it('should load attachments', (done) => {
    component.kommentar = createKommentarResource([KommentarLinkRel.ATTACHMENTS]);
    kommentarService.getAttachments.mockReturnValue(of(createStateResource(createBinaryFileListResource())));

    component.ngOnChanges();

    component.attachments$.subscribe((binaryFileResources) => {
      expect(binaryFileResources.length).toBe(10);
      done();
    });
  });

  it('should call kommentarService', () => {
    const kommentarResource = createKommentarResource([KommentarLinkRel.ATTACHMENTS]);
    component.kommentar = kommentarResource;
    kommentarService.getAttachments.mockReturnValue(of(createStateResource(createBinaryFileListResource())));

    component.ngOnChanges();

    expect(kommentarService.getAttachments).toHaveBeenCalledWith(kommentarResource);
  });

  describe('submit', () => {
    it('should submit form', () => {
      formService.submit = jest.fn().mockReturnValue(EMPTY);

      component.submit();

      expect(formService.submit).toHaveBeenCalled();
    });

    it('should set submit in progress', () => {
      const command: StateResource<CommandResource> = createSuccessfullyDoneCommandStateResource();
      formService.submit = jest.fn().mockReturnValue(of(command));

      component.submit();

      expect(component.submitInProgress$).toBeObservable(singleColdCompleted(command));
    });

    it('should clear uploaded attachments', () => {
      formService.submit = jest.fn().mockReturnValue(of(createSuccessfullyDoneCommandStateResource()));

      component.submit();
      component.submitInProgress$.subscribe();

      expect(kommentarService.clearUploadedFiles).toHaveBeenCalled();
    });

    it('should NOT clear uploaded attachments on loading', () => {
      formService.submit = jest.fn().mockReturnValue(of(createEmptyStateResource(true)));

      component.submit();
      component.submitInProgress$.subscribe();

      expect(kommentarService.clearUploadedFiles).not.toHaveBeenCalled();
    });

    it('should NOT clear uploaded attachments on error', () => {
      formService.submit = jest.fn().mockReturnValue(of(createErrorStateResource(createProblemDetail())));

      component.submit();
      component.submitInProgress$.subscribe();

      expect(kommentarService.clearUploadedFiles).not.toHaveBeenCalled();
    });
  });

  describe('onCancel', () => {
    beforeEach(() => {
      component.cancel.emit = jest.fn();
    });

    it('should clear uploaded attachments', () => {
      component.onCancel();

      expect(kommentarService.clearUploadedFiles).toHaveBeenCalled();
    });

    it('should emit', () => {
      component.onCancel();

      expect(component.cancel.emit).toHaveBeenCalled();
    });
  });

  describe('template', () => {
    describe('cancel button', () => {
      it('should exists', () => {
        existsAsHtmlElement(fixture, cancelButtonTestId);
      });

      describe('output', () => {
        describe('clickEmitter', () => {
          it('should call handler', () => {
            component.onCancel = jest.fn();

            triggerEvent({ fixture, elementSelector: cancelButtonTestId, name: 'clickEmitter' });

            expect(component.onCancel).toHaveBeenCalled();
          });
        });
      });
    });
  });
});
