/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { KommentarLinkRel, KommentarResource, KommentarService } from '@alfa-client/kommentar-shared';
import { ConvertForDataTestPipe, FormatDateWithTimePipe, HasLinkPipe } from '@alfa-client/tech-shared';
import { getElementFromFixture, mock } from '@alfa-client/test-utils';
import { UserProfileInKommentarContainerComponent } from '@alfa-client/user-profile';
import { registerLocaleData } from '@angular/common';
import localeDe from '@angular/common/locales/de';
import { ComponentFixture, TestBed } from '@angular/core/testing';
import { faker } from '@faker-js/faker/.';
import { expect } from '@jest/globals';
import { getUrl } from '@ngxp/rest';
import { createKommentarResource } from 'libs/kommentar-shared/test/kommentar';
import { getDataTestClassOf } from 'libs/tech-shared/test/data-test';
import { MockComponent } from 'ng-mocks';
import { KommentarFormComponent } from '../../kommentar-form/kommentar-form.component';
import { KommentarListItemInVorgangComponent } from './kommentar-list-item-in-vorgang.component';

registerLocaleData(localeDe);

describe('KommentarListItemInVorgangComponent', () => {
  let component: KommentarListItemInVorgangComponent;
  let fixture: ComponentFixture<KommentarListItemInVorgangComponent>;

  const userProfile: string = getDataTestClassOf('kommentar-created-by');
  const kommentarService = mock(KommentarService);

  beforeEach(async () => {
    await TestBed.configureTestingModule({
      declarations: [
        HasLinkPipe,
        KommentarListItemInVorgangComponent,
        ConvertForDataTestPipe,
        FormatDateWithTimePipe,
        MockComponent(KommentarFormComponent),
        MockComponent(UserProfileInKommentarContainerComponent),
      ],
      providers: [
        {
          provide: KommentarService,
          useValue: kommentarService,
        },
      ],
    });
  });

  beforeEach(() => {
    fixture = TestBed.createComponent(KommentarListItemInVorgangComponent);
    component = fixture.componentInstance;
    component.kommentar = createKommentarResource();
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  describe('set currently edited', () => {
    const kommentar: KommentarResource = createKommentarResource();

    it('should set edit mode to false on same currently edited resource uri', () => {
      component.editMode = false;
      component.kommentar = kommentar;

      component.currentlyEdited = getUrl(kommentar);

      expect(component.editMode).toBe(false);
    });

    it('should set edit mode to false on different currently edited resource uri', () => {
      component.editMode = true;
      component.kommentar = kommentar;

      component.currentlyEdited = faker.internet.url();

      expect(component.editMode).toBe(false);
    });

    it('should set edit mode to true', () => {
      component.editMode = true;
      component.kommentar = kommentar;

      component.currentlyEdited = getUrl(kommentar);

      expect(component.editMode).toBe(true);
    });
  });

  describe('user profile', () => {
    it('should be visible on existing link', () => {
      component.kommentar = createKommentarResource([KommentarLinkRel.CREATED_BY]);
      fixture.detectChanges();

      const element = getElementFromFixture(fixture, userProfile);

      expect(element).toBeInstanceOf(HTMLElement);
    });

    it('should be hidden if link is missing', () => {
      component.kommentar = createKommentarResource();
      fixture.detectChanges();

      const element = getElementFromFixture(fixture, userProfile);

      expect(element).not.toBeInstanceOf(HTMLElement);
    });
  });

  describe('edit', () => {
    const kommentar: KommentarResource = createKommentarResource();

    it('should change editMode', () => {
      component.kommentar = createKommentarResource([KommentarLinkRel.EDIT]);

      component.edit();

      expect(component.editMode).toBeTruthy();
    });

    it('should NOT change editMode', () => {
      component.kommentar = kommentar;

      component.edit();

      expect(component.editMode).toBeFalsy();
    });

    it('should set currently edited kommetar uri', () => {
      component.kommentar = kommentar;

      component.edit();

      expect(kommentarService.setCurrentlyEdited).toHaveBeenCalledWith(getUrl(kommentar));
    });
  });
});
