/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { BinaryFileService } from '@alfa-client/binary-file-shared';
import { CommandResource, CommandService } from '@alfa-client/command-shared';
import { NavigationService } from '@alfa-client/navigation-shared';
import { createEmptyStateResource, createStateResource, StateResource } from '@alfa-client/tech-shared';
import { Mock, mock } from '@alfa-client/test-utils';
import { SnackBarService } from '@alfa-client/ui';
import { VorgangService, VorgangWithEingangResource } from '@alfa-client/vorgang-shared';
import { TestBed } from '@angular/core/testing';
import { MatDialog } from '@angular/material/dialog';
import { expect } from '@jest/globals';
import { CommandLinkRel } from 'libs/command-shared/src/lib/command.linkrel';
import { createCommandErrorResource, createCommandResource } from 'libs/command-shared/test/command';
import { createVorgangWithEingangResource } from 'libs/vorgang-shared/test/vorgang';
import { BehaviorSubject, of } from 'rxjs';
import { createPostfachFeatures, createPostfachMail, createPostfachMailListResource, createPostfachMailResource, createPostfachSettings, } from '../../test/postfach';
import { PostfachFacade } from './+state/postfach.facade';
import { PostfachMailLinkRel, PostfachMailListLinkRel } from './postfach.linkrel';
import { PostfachMessages } from './postfach.message';
import { PostfachMail, PostfachMailListResource, PostfachMailResource } from './postfach.model';
import { PostfachRepository } from './postfach.repository';
import { PostfachService } from './postfach.service';
import { createSendPostfachMailCommand } from './postfach.util';

describe('PostfachService', () => {
  let service: PostfachService;

  let repository: Mock<PostfachRepository>;
  const commandService: Mock<CommandService> = mock(CommandService);
  const navigationService: Mock<NavigationService> = mock(NavigationService);
  const vorgangService: Mock<VorgangService> = mock(VorgangService);
  const snackbarService: Mock<SnackBarService> = mock(SnackBarService);
  const dialog: Mock<MatDialog> = <Mock<MatDialog>>{ closeAll: jest.fn() };
  const postfachFacade: Mock<PostfachFacade> = mock(PostfachFacade);
  const binaryFileService: Mock<BinaryFileService> = mock(BinaryFileService);

  const urlChangedParams = {};

  beforeEach(() => {
    navigationService.urlChanged = jest.fn();
    navigationService.urlChanged.mockReturnValue(of(urlChangedParams));

    repository = mock(PostfachRepository);

    TestBed.configureTestingModule({
      providers: [
        { provide: CommandService, useValue: commandService },
        { provide: NavigationService, useValue: navigationService },
        { provide: VorgangService, useValue: vorgangService },
        { provide: SnackBarService, useValue: snackbarService },
        { provide: MatDialog, useValue: dialog },
        { provide: PostfachFacade, useValue: postfachFacade },
        { provide: BinaryFileService, useValue: binaryFileService },
        { provide: PostfachRepository, useValue: repository },
        PostfachService,
      ],
    });

    service = TestBed.inject(PostfachService);
  });

  it('should be created', () => {
    expect(service).toBeTruthy();
  });

  describe('send message', () => {
    const postfachMail: PostfachMail = createPostfachMail();
    const commandStateResource: StateResource<CommandResource> = createStateResource(createCommandResource());

    beforeEach(() => {
      commandService.createCommand.mockReturnValue(of(commandStateResource));
      vorgangService.getPendingSendPostfachMailCommand.mockReturnValue(of(commandStateResource));
    });

    it('should call vorgang service and set command on loading', () => {
      service.sendMail(postfachMail);

      expect(vorgangService.setPendingSendPostfachMailSingleCommandLoading).toHaveBeenCalled();
    });

    it('should call command service', () => {
      service.sendMail(postfachMail);

      expect(commandService.createCommand).toHaveBeenCalled();
    });

    it('should call vorgang service and get command', () => {
      service.sendMail(postfachMail);

      expect(vorgangService.getPendingSendPostfachMailCommand).toHaveBeenCalled();
    });

    describe('reload postfach list', () => {
      beforeEach(() => {
        service.refreshPostfachMailList = jest.fn();
      });

      it('should execute on command is done', () => {
        const commandStateResource: StateResource<CommandResource> = createStateResource(
          createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]),
        );
        commandService.createCommand.mockReturnValue(of(commandStateResource));

        service.sendMail(postfachMail);

        expect(service.refreshPostfachMailList).toHaveBeenCalledWith(commandStateResource);
      });

      it('should not execute on pending command', () => {
        commandService.createCommand.mockReturnValue(of(createStateResource(createCommandResource([CommandLinkRel.UPDATE]))));

        service.sendMail(postfachMail);

        expect(service.refreshPostfachMailList).not.toHaveBeenCalled();
      });
    });
  });

  describe('sendExistingNachricht', () => {
    const postfachNachricht: PostfachMailResource = createPostfachMailResource();
    const nachricht: PostfachMail = createPostfachMail();

    const commandStateResource: StateResource<CommandResource> = createStateResource(createCommandResource());

    it('should call doSendNachricht', () => {
      service.doSendNachricht = jest.fn();
      commandService.createCommand.mockReturnValue(of(commandStateResource));
      vorgangService.getPendingSendPostfachMailCommand.mockReturnValue(of(commandStateResource));

      service.sendExistingMail(postfachNachricht, nachricht);

      expect(service.doSendNachricht).toHaveBeenCalledWith(
        postfachNachricht,
        PostfachMailLinkRel.SEND,
        createSendPostfachMailCommand(nachricht),
      );
    });
  });

  describe('resend message', () => {
    const postfachMailResource: PostfachMailResource = createPostfachMailResource();

    beforeEach(() => {
      service.handleSendPostfachMailIsDone = jest.fn();

      commandService.createCommand.mockReturnValue(of(createStateResource(createCommandResource())));
      vorgangService.getPendingSendPostfachMailCommand.mockReturnValue(of(createStateResource(createCommandResource())));
    });

    it('should call vorgang service and set command on loading', () => {
      service.resendMail(postfachMailResource);

      expect(vorgangService.setPendingSendPostfachMailSingleCommandLoading).toHaveBeenCalled();
    });

    it('should call command service', () => {
      service.resendMail(postfachMailResource);

      expect(commandService.createCommand).toHaveBeenCalled();
    });

    it('should call vorgang service and get command', () => {
      service.resendMail(postfachMailResource);

      expect(vorgangService.getPendingSendPostfachMailCommand).toHaveBeenCalled();
    });

    it('should NOT call handleSendPostfachMailIsDone if command is pending', () => {
      commandService.createCommand.mockReturnValue(of(createStateResource(createCommandResource([CommandLinkRel.UPDATE]))));
      service.resendMail(postfachMailResource);

      expect(service.handleSendPostfachMailIsDone).not.toHaveBeenCalled();
    });

    it('should call handleSendPostfachMailIsDone if command is done', () => {
      const commandResponse: StateResource<CommandResource> = createStateResource(
        createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]),
      );
      commandService.createCommand.mockReturnValue(of(commandResponse));

      service.resendMail(postfachMailResource);

      expect(service.handleSendPostfachMailIsDone).toHaveBeenCalledWith(commandResponse);
    });
  });

  describe('handleSendNachrichtCommand', () => {
    const commandStateResource: StateResource<CommandResource> = createStateResource(createCommandResource());

    beforeEach(() => {
      service.commandIsDone = jest.fn();
    });

    it('should call vorgang service', () => {
      service.handleSendNachrichtCommand(commandStateResource);

      expect(vorgangService.setPendingSendPostfachMailCommand).toHaveBeenCalledWith(commandStateResource);
    });

    it('should call commandIsDone', () => {
      service.handleSendNachrichtCommand(commandStateResource);

      expect(service.commandIsDone).toHaveBeenCalledWith(commandStateResource);
    });
  });

  describe('commandIsDone', () => {
    beforeEach(() => {
      service.handleSendPostfachMailIsDone = jest.fn();
    });

    it('should call handleSendPostfachMailIsDone', () => {
      const commandStateResource: StateResource<CommandResource> = createStateResource(
        createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]),
      );

      service.commandIsDone(commandStateResource);

      expect(service.handleSendPostfachMailIsDone).toHaveBeenCalledWith(commandStateResource);
    });

    describe('on command is done', () => {
      it('should call snackbar service', () => {
        const commandStateResource: StateResource<CommandResource> = createStateResource({
          ...createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]),
          errorMessage: null,
        });

        service.commandIsDone(commandStateResource);

        expect(snackbarService.show).toHaveBeenCalledWith(commandStateResource.resource, PostfachMessages.SEND_SUCCESSFUL);
      });
    });

    describe('on command is done with error', () => {
      it('should call snackbar service', () => {
        const commandStateResource: StateResource<CommandResource> = createStateResource(
          createCommandErrorResource([CommandLinkRel.EFFECTED_RESOURCE]),
        );

        service.commandIsDone(commandStateResource);

        expect(snackbarService.showError).toHaveBeenCalledWith(PostfachMessages.SEND_FAILED);
      });
    });
  });

  describe('listenToNavigation', () => {
    it('should call usnsubcribeToNavigation', () => {
      service.listenToNavigation();

      expect(navigationService.urlChanged).toHaveBeenCalled();
    });

    it('should call navigation service', () => {
      service.unsubscribeToNavigation = jest.fn();

      service.listenToNavigation();

      expect(service.unsubscribeToNavigation).toHaveBeenCalled();
    });

    it('should call onNavigation', () => {
      service.onNavigation = jest.fn();

      service.listenToNavigation();

      expect(service.onNavigation).toHaveBeenCalledWith(urlChangedParams);
    });
  });

  describe('onNavigation', () => {
    describe('to vorgang list', () => {
      beforeEach(() => {
        service.setPollingFalse = jest.fn();
        service.clearPostfachMailList = jest.fn();
        service.unsubscribe = jest.fn();
      });

      it('should set polling false', () => {
        service.onNavigation({});

        expect(service.setPollingFalse).toHaveBeenCalled();
      });

      it('should call clearPostfachMailList', () => {
        service.onNavigation({});

        expect(service.clearPostfachMailList).toHaveBeenCalled();
      });

      it('should call closeOpenDialogs', () => {
        service.onNavigation({});

        expect(dialog.closeAll).toHaveBeenCalled();
      });

      it('should call unsubscribe', () => {
        service.onNavigation({});

        expect(service.unsubscribe).toHaveBeenCalled();
      });
    });

    describe('to vorgang detail page', () => {
      it('should set postfach nachrichten list to reload', () => {
        service.postfachMailList$.next(createEmptyStateResource());

        service.onNavigation({ vorgangWithEingangUrl: 'X' });

        expect(service.postfachMailList$.value.reload).toEqual(true);
      });
    });

    describe('to postfach page', () => {
      beforeEach(() => {
        service.resetHasNewPostfachNachrichten = jest.fn();
        navigationService.isPostfachPage.mockReturnValue(true);
      });

      it('should call navigationService', () => {
        service.onNavigation({});

        expect(navigationService.isPostfachPage).toHaveBeenCalled();
      });

      it('should reset hasNewPostfachNachrichten', () => {
        navigationService.isPostfachPage.mockReturnValue(true);

        service.onNavigation({});

        expect(service.resetHasNewPostfachNachrichten).toHaveBeenCalled();
      });
    });
  });

  describe('getPendingSendPostfachMailCommand', () => {
    beforeEach(() => {
      service.unsubscribe = jest.fn();
      service.pollSendPostfachMailCommand = jest.fn();

      commandService.pollCommand.mockReturnValue(of(createStateResource(createCommandResource())));
      vorgangService.getPendingSendPostfachMailCommand.mockReturnValue(of(createStateResource(createCommandResource())));
    });

    it('should call navigation service', () => {
      service.getPendingSendPostfachMailCommand();

      expect(navigationService.urlChanged).toHaveBeenCalled();
    });
  });

  describe('resetHasNewPostfachNachrichten', () => {
    const postfachListNachrichtenResource: PostfachMailListResource = createPostfachMailListResource();

    it('should call doResetHasNewPostfachNachrichten', () => {
      service.getPostfachMailListByVorgang = jest.fn();
      service.doResetHasNewPostfachNachrichten = jest.fn();
      (<any>service.getPostfachMailListByVorgang).mockReturnValue(of(postfachListNachrichtenResource));

      service.resetHasNewPostfachNachrichten();

      expect(service.doResetHasNewPostfachNachrichten).toHaveBeenCalled();
    });
  });

  describe('doResetHasNewPostfachNachrichten', () => {
    describe('on existing link', () => {
      const postfachNachrichtenListResource: PostfachMailListResource = createPostfachMailListResource([
        PostfachMailListLinkRel.RESET_HAS_NEW_POSTFACH_NACHRICHT,
      ]);

      beforeEach(() => {
        service.postfachMailList$.next(createStateResource(postfachNachrichtenListResource));
        repository.resetHasNewPostfachNachrichten.mockReturnValue(of(postfachNachrichtenListResource));
      });

      it('should call repository if link exists', () => {
        service.doResetHasNewPostfachNachrichten();

        expect(repository.resetHasNewPostfachNachrichten).toHaveBeenCalledWith(postfachNachrichtenListResource);
      });
    });

    it('should NOT call repository if link not exists', () => {
      const postfachNachrichtenListResource: PostfachMailListResource = createPostfachMailListResource();
      service.postfachMailList$.next(createStateResource(postfachNachrichtenListResource));

      service.doResetHasNewPostfachNachrichten();

      expect(repository.resetHasNewPostfachNachrichten).not.toHaveBeenCalled();
    });
  });

  describe('getPostfachMailListByVorgang', () => {
    const postfachMailList: PostfachMailListResource = createPostfachMailListResource();

    beforeEach(() => {
      repository.loadPostfachMailList.mockReturnValue(of(postfachMailList));
      service.setPostfachMailListLoading = jest.fn();
      service.setPostfachMailList = jest.fn();
      vorgangService.getVorgangWithEingang.mockReturnValue(of(createStateResource(createVorgangWithEingangResource())));
      service.postfachMailList$.next(createEmptyStateResource());
    });

    it('should set loading to true', () => {
      service.getPostfachMailListByVorgang();

      expect(service.setPostfachMailListLoading).toHaveBeenCalled();
    });

    it('should call vorgang service', () => {
      service.getPostfachMailListByVorgang();

      expect(vorgangService.getVorgangWithEingang).toHaveBeenCalled();
    });

    it('should call repository', () => {
      service.getPostfachMailListByVorgang().subscribe();

      expect(repository.loadPostfachMailList).toHaveBeenCalled();
    });

    it('should set loading to false', () => {
      service.getPostfachMailListByVorgang();

      expect(service.setPostfachMailList).toHaveBeenCalled();
    });
  });

  describe('handleSendPostfachMailIsDone', () => {
    const stateResource: StateResource<CommandResource> = createStateResource(createCommandResource());

    beforeEach(() => {
      service.setPollingFalse = jest.fn();
      service.setPostfachMailList = jest.fn();

      commandService.getEffectedResource.mockReturnValue(of(createCommandResource()));
    });

    it('should call command service to get effected resource', () => {
      service.handleSendPostfachMailIsDone(stateResource);

      expect(commandService.getEffectedResource).toHaveBeenCalledWith(stateResource.resource);
    });

    it('should set postfachMailList', () => {
      service.handleSendPostfachMailIsDone(stateResource);

      expect(service.setPostfachMailList).toHaveBeenCalled();
    });

    it('should set polling to false', () => {
      service.handleSendPostfachMailIsDone(stateResource);

      expect(service.setPollingFalse).toHaveBeenCalled();
    });
  });

  describe('isDownloadPdfInProgress', () => {
    const vorgang: VorgangWithEingangResource = createVorgangWithEingangResource();

    it('should call facade to download pdf', (done) => {
      vorgangService.getVorgangWithEingang.mockReturnValue(of(createStateResource(vorgang)));
      postfachFacade.isDownloadPdfInProgress.mockReturnValue(of(true));

      service.isDownloadPdfInProgress().subscribe(() => {
        expect(postfachFacade.downloadPdf).toHaveBeenCalledWith(vorgang);
        done();
      });
    });

    describe('combine latest', () => {
      beforeEach(() => {
        vorgangService.getVorgangWithEingang.mockReturnValue(of(vorgang));
        postfachFacade.isDownloadPdfInProgress = jest.fn().mockReturnValue(true);
      });

      it('should call facade to get isDownloadPdfInProgress', () => {
        service.isDownloadPdfInProgress();

        expect(postfachFacade.isDownloadPdfInProgress).toHaveBeenCalled();
      });

      it('should call vorgang service to get vorgang', () => {
        service.isDownloadPdfInProgress();

        expect(vorgangService.getVorgangWithEingang).toHaveBeenCalled();
      });
    });
  });

  describe('downloadPdf', () => {
    it('should call facade', () => {
      service.downloadPdf();

      expect(postfachFacade.startDownloadPdf).toHaveBeenCalled();
    });
  });

  describe('getFeatures', () => {
    it('should return features by list stateResource', (done) => {
      service.postfachMailList$ = new BehaviorSubject(createStateResource(createPostfachMailListResource()));

      service.getFeatures().subscribe((features) => {
        expect(features).toEqual(createPostfachFeatures());
        done();
      });
    });
  });

  describe('getSettings', () => {
    it('should return settings by list stateResource', (done) => {
      service.postfachMailList$ = new BehaviorSubject(createStateResource(createPostfachMailListResource()));

      service.getSettings().subscribe((settings) => {
        expect(settings).toEqual(createPostfachSettings());
        done();
      });
    });
  });

  describe('close open dialogs', () => {
    beforeEach(() => {
      service.clearUploadedFiles = jest.fn();
    });

    it('should clear uploaded files', () => {
      service.closeOpenDialogs();

      expect(service.clearUploadedFiles).toHaveBeenCalled();
    });
  });
});
